import itertools as it
import random

import opytional as opyt
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
def test_CalcRanksSinceLastRetainedCommonalitytWith_specimen(
    retention_policy, differentia_width
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_ranks_since_last_retained_commonality_with(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column)
    ) == hstrat.calc_ranks_since_last_retained_commonality_with(column, column)

    assert hstrat.calc_ranks_since_last_retained_commonality_with(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(child1)
    ) == hstrat.calc_ranks_since_last_retained_commonality_with(column, child1)

    assert hstrat.calc_ranks_since_last_retained_commonality_with(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.calc_ranks_since_last_retained_commonality_with(child1, child2)

    child1.DepositStrata(10)
    assert hstrat.calc_ranks_since_last_retained_commonality_with(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.calc_ranks_since_last_retained_commonality_with(child1, child2)


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy_2b,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    ordered_store,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_ranks_since_last_retained_commonality_with(
                first, second
            ) == hstrat.calc_ranks_since_last_retained_commonality_with(
                second, first
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_validity(retention_policy, ordered_store):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for generation in range(100):
        for first, second in it.combinations(population, 2):
            rslcw = hstrat.calc_ranks_since_last_retained_commonality_with(
                first, second
            )
            if rslcw is not None:
                assert 0 <= rslcw <= generation

            rsfdw = hstrat.calc_ranks_since_first_retained_disparity_with(
                first, second
            )
            if rsfdw is not None:
                assert -1 <= rsfdw <= generation

            assert hstrat.calc_ranks_since_mrca_bounds_with(
                first,
                second,
                prior="arbitrary",
            ) is None or hstrat.calc_ranks_since_mrca_bounds_with(
                first,
                second,
                prior="arbitrary",
            ) == (
                opyt.or_value(rsfdw, -1) + 1,
                rslcw + 1,
            )
            if rslcw is not None and rsfdw is not None:
                assert rsfdw < rslcw

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy1",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "retention_policy2",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_no_mrca(
    retention_policy1,
    retention_policy2,
    ordered_store,
):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy1,
    )
    second = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy2,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_ranks_since_last_retained_commonality_with(
                first, second
            )
            is None
        )
        assert (
            hstrat.calc_ranks_since_last_retained_commonality_with(
                second, first
            )
            is None
        )
        first.DepositStratum()
        second.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_no_divergence(retention_policy, ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_ranks_since_last_retained_commonality_with(
                column, column
            )
            == 0
        )
        column.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_partial_even_divergence(retention_policy, ordered_store):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    first.DepositStrata(100)

    second = first.Clone()

    first.DepositStratum()
    second.DepositStratum()

    for generation in range(101, 200):
        assert (
            generation - 101
            < hstrat.calc_ranks_since_last_retained_commonality_with(
                first, second
            )
            <= generation
        )

        first.DepositStratum()
        second.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_scenario_partial_uneven_divergence(retention_policy, ordered_store):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    first.DepositStrata(100)

    second = first.Clone()

    first.DepositStratum()

    for generation in range(101, 200):
        assert (
            generation - 101
            < hstrat.calc_ranks_since_last_retained_commonality_with(
                first, second
            )
            <= generation
        )
        assert (
            0
            <= hstrat.calc_ranks_since_last_retained_commonality_with(
                second, first
            )
            <= 100
        )

        first.DepositStratum()

    second.DepositStratum()

    for generation in range(101, 200):
        assert (
            100
            <= hstrat.calc_ranks_since_last_retained_commonality_with(
                first, second
            )
            <= 200
        )
        assert (
            0
            <= hstrat.calc_ranks_since_last_retained_commonality_with(
                second, first
            )
            <= generation
        )

        second.DepositStratum()


@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
def test_CalcRanksSinceLastRetainedCommonalityWith1(differentia_width):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_width,
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_ranks_since_last_retained_commonality_with(
                    c1,
                    c2,
                )
                == c1.GetNumStrataDeposited() - column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_ranks_since_last_retained_commonality_with(
                    c2,
                    c1,
                )
                == c2.GetNumStrataDeposited() - column.GetNumStrataDeposited()
            )
        elif differentia_width == 1:
            assert (
                hstrat.calc_ranks_since_last_retained_commonality_with(
                    c1,
                    c2,
                )
                > c1.GetNumStrataDeposited() - column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_ranks_since_last_retained_commonality_with(
                    c2,
                    c1,
                )
                > c1.GetNumStrataDeposited() - column.GetNumStrataDeposited()
            )

    for c in [column, offspring1, offspring2]:
        assert (
            hstrat.calc_ranks_since_last_retained_commonality_with(c, c, 0.8)
            == c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.2,
            )
            - 1
        )
        assert (
            hstrat.calc_ranks_since_last_retained_commonality_with(c, c, 0.99)
            == c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.01,
            )
            - 1
        )

    offspring1.DepositStrata(100)
    offspring2.DepositStrata(100)

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_ranks_since_last_retained_commonality_with(
                    c1,
                    c2,
                )
                == c1.GetNumStrataDeposited() - column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_ranks_since_last_retained_commonality_with(
                    c2,
                    c1,
                )
                == c2.GetNumStrataDeposited() - column.GetNumStrataDeposited()
            )
        elif differentia_width == 1:
            assert hstrat.calc_ranks_since_last_retained_commonality_with(
                c1, c2, 0.999999
            ) > hstrat.calc_ranks_since_last_retained_commonality_with(
                c1, c2, 0.8
            )

    for c in [column, offspring1, offspring2]:
        assert (
            hstrat.calc_ranks_since_last_retained_commonality_with(c, c, 0.8)
            == c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.2,
            )
            - 1
        )
        assert (
            hstrat.calc_ranks_since_last_retained_commonality_with(c, c, 0.99)
            == c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.01,
            )
            - 1
        )


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_ranks_since_last_retained_commonality_with(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.calc_ranks_since_last_retained_commonality_with(a, b)



import random
from typing import Any


class RngStateContext:
    """A context manager that temporarily reseeds the `random` module random
    number generator (RNG) and restores the original RNG state upon exiting the
    scope.
    """

    def __init__(self, seed: int):
        """Constructs an RngStateContext with the specified seed value.

        Parameters:
        -----------
        seed: int
            The seed value to use for reseeding the RNG temporarily.
        """
        self._saved_state = None
        self.seed = seed

    def __enter__(self) -> None:
        """Enters the context manager and saves the current state of the RNG to
        _saved_state, then reseeds the RNG with the specified seed value."""
        self._saved_state = random.getstate()
        random.seed(self.seed)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """Exits the context manager and restores the RNG state to the value
        saved in _saved_state."""
        random.setstate(self._saved_state)



"""Functors to specify property policies should be parameterized for."""

from ._MrcaUncertaintyAbsExactEvaluator import MrcaUncertaintyAbsExactEvaluator
from ._MrcaUncertaintyAbsUpperBoundEvaluator import (
    MrcaUncertaintyAbsUpperBoundEvaluator,
)
from ._MrcaUncertaintyRelExactEvaluator import MrcaUncertaintyRelExactEvaluator
from ._MrcaUncertaintyRelUpperBoundEvaluator import (
    MrcaUncertaintyRelUpperBoundEvaluator,
)
from ._NumStrataRetainedExactEvaluator import NumStrataRetainedExactEvaluator
from ._NumStrataRetainedUpperBoundEvaluator import (
    NumStrataRetainedUpperBoundEvaluator,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "MrcaUncertaintyAbsExactEvaluator",
    "MrcaUncertaintyAbsUpperBoundEvaluator",
    "MrcaUncertaintyRelExactEvaluator",
    "MrcaUncertaintyRelUpperBoundEvaluator",
    "NumStrataRetainedExactEvaluator",
    "NumStrataRetainedUpperBoundEvaluator",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



"""Tools for perfect phylogenetic tracking."""

from ._DecantingPhyloTracker import DecantingPhyloTracker
from ._GarbageCollectingPhyloTracker import GarbageCollectingPhyloTracker
from ._PerfectBacktrackHandle import PerfectBacktrackHandle
from ._compile_perfect_backtrack_phylogeny import (
    compile_perfect_backtrack_phylogeny,
)
from ._compile_phylogeny_from_lineage_iters import (
    compile_phylogeny_from_lineage_iters,
)
from ._iter_perfect_backtrack_lineage import iter_perfect_backtrack_lineage

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "compile_perfect_backtrack_phylogeny",
    "compile_phylogeny_from_lineage_iters",
    "DecantingPhyloTracker",
    "GarbageCollectingPhyloTracker",
    "iter_perfect_backtrack_lineage",
    "PerfectBacktrackHandle",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_is_topologically_sorted,
    alifestd_parse_ancestor_ids,
    alifestd_to_working_format,
    alifestd_topological_sort,
    alifestd_try_add_ancestor_id_col,
    swap_rows_and_indices,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_topological_sort,
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_is_topologically_sorted_empty(phylogeny_df, apply):
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df.sort_values("id", ascending=True, inplace=True)

    assert alifestd_is_topologically_sorted(phylogeny_df.iloc[-1:0, :])


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_topological_sort,
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_is_topologically_sorted_singleton(phylogeny_df, apply):
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df.sort_values("id", ascending=True, inplace=True)

    assert alifestd_is_topologically_sorted(phylogeny_df.iloc[0:1, :])


def test_alifestd_is_topologically_sorted_tworoots():

    phylo1 = pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    phylo1.sort_values("id", ascending=True, inplace=True)

    phylo2 = pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
    phylo2.sort_values("id", ascending=True, inplace=True)

    assert alifestd_is_topologically_sorted(
        pd.concat(
            [
                phylo1.iloc[0:1, :],
                phylo2.iloc[0:1, :],
            ]
        )
    )


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_is_topologically_sorted_twolineages_true(phylogeny_df):

    phylogeny_df.sort_values("id", ascending=True, inplace=True)
    phylogeny_df.reset_index(inplace=True)

    aggregated_df = alifestd_aggregate_phylogenies(
        [
            phylogeny_df.copy(),
            phylogeny_df.copy(),
        ]
    )

    lineage1 = aggregated_df[len(phylogeny_df) :]
    lineage2 = aggregated_df[: len(phylogeny_df)]

    assert alifestd_is_topologically_sorted(
        pd.concat(
            [
                lineage1,
                lineage2,
            ]
        )
    )
    assert alifestd_is_topologically_sorted(
        pd.concat(
            [
                lineage1,
                lineage2,
            ]
        ).sort_index()
    )


def test_alifestd_is_topologically_sorted_twolineages_true_sexual():
    assert alifestd_is_topologically_sorted(
        pd.read_csv(f"{assets_path}/example-standard-toy-sexual-phylogeny.csv")
    )


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_topological_sort,
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_is_topologically_sorted_twolineages_false(
    phylogeny_df, apply
):
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df.sort_values("id", ascending=True, inplace=True)
    phylogeny_df.reset_index(inplace=True)
    max_id = phylogeny_df["id"].max()

    lineage1 = phylogeny_df.copy()
    lineage2 = phylogeny_df.copy()
    lineage2["id"] += max_id + 1
    lineage2["ancestor_list"] = lineage2["ancestor_list"].apply(
        lambda ancestor_list_str: str(
            [
                ancestor_id + max_id + 1
                for ancestor_id in alifestd_parse_ancestor_ids(
                    ancestor_list_str
                )
            ]
        )
    )
    assert not alifestd_is_topologically_sorted(
        pd.concat(
            [
                lineage1.iloc[::-1, :],
                lineage2,
            ]
        )
    )
    assert not alifestd_is_topologically_sorted(
        pd.concat(
            [
                lineage1.iloc[:10:-1, :],
                lineage2,
                lineage1.iloc[10::-1, :],
            ]
        )
    )
    assert not alifestd_is_topologically_sorted(
        pd.concat(
            [
                lineage1.iloc[10::, :],
                lineage2,
                lineage1.iloc[:10:, :],
            ]
        )
    )


def test_alifestd_is_topologically_sorted_twolineages_false_sexual():
    df = pd.read_csv(
        f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
    )
    assert not alifestd_is_topologically_sorted(df.iloc[::-1])


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_topological_sort,
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_is_topologically_sorted_true(phylogeny_df, apply):
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df.sort_values("id", ascending=True, inplace=True)

    phylogeny_df_ = phylogeny_df.copy()
    assert alifestd_is_topologically_sorted(phylogeny_df)

    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_topological_sort,
        alifestd_to_working_format,
        alifestd_try_add_ancestor_id_col,
        lambda x: x,
    ],
)
def test_alifestd_is_topologically_sorted_false(phylogeny_df, apply):
    phylogeny_df = apply(phylogeny_df)

    phylogeny_df = phylogeny_df.copy()
    phylogeny_df.sort_values("id", ascending=False, inplace=True)
    phylogeny_df.set_index("id", drop=False, inplace=True)

    phylogeny_df_ = phylogeny_df.copy()
    assert not alifestd_is_topologically_sorted(phylogeny_df)

    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)

    # reverse dataframe
    phylogeny_df = phylogeny_df.iloc[::-1]
    assert alifestd_is_topologically_sorted(phylogeny_df)

    # one-by-one transpositions
    for idx, row in phylogeny_df.iterrows():
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            phylogeny_df_ = swap_rows_and_indices(
                phylogeny_df, idx, ancestor_id
            )

            assert not alifestd_is_topologically_sorted(phylogeny_df_)

    # cumulative transpositions in random order
    for idx, row in phylogeny_df.sample(frac=1).iterrows():
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            assert ancestor_id in phylogeny_df_.index
            phylogeny_df = swap_rows_and_indices(
                phylogeny_df, idx, ancestor_id
            )

            assert not alifestd_is_topologically_sorted(phylogeny_df)



import itertools as it
import numbers

from iterpop import iterpop as ip
import numpy as np
import pytest

from hstrat._auxiliary_lib import pairwise
from hstrat.hstrat import geom_seq_nth_root_tapered_algo


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        pytest.param(1, marks=pytest.mark.heavy_3a),
        2,
        3,
        7,
        9,
        pytest.param(42, marks=pytest.mark.heavy_2a),
        pytest.param(97, marks=pytest.mark.heavy_2a),
        pytest.param(100, marks=pytest.mark.heavy_2a),
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        pytest.param(1, marks=pytest.mark.heavy_3b),
        2,
        5,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        pytest.param(
            it.chain(
                range(10**4),
                np.logspace(10, 32, num=50, base=2, dtype="int"),
            ),
            marks=pytest.mark.heavy_3c,
        ),
        (i for i in range(10) for __ in range(2)),
        (10 - i for i in range(10) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**2,
            size=10,
        ),
        pytest.param(
            np.random.default_rng(1).integers(
                low=0,
                high=2**32,
                size=10,
            ),
            marks=pytest.mark.heavy_2b,
        ),
    ],
)
def test_only_dwindling_over_time(degree, interspersal, time_sequence):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec)
    for num_strata_deposited in time_sequence:
        for which in (
            instance,
            geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec),
        ):
            cur_set = {
                *which(
                    policy,
                    num_strata_deposited,
                )
            }
            next_set = {
                *which(
                    policy,
                    num_strata_deposited + 1,
                )
            }
            assert cur_set.issuperset(next_set - {num_strata_deposited})


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        pytest.param(1, marks=pytest.mark.heavy_3a),
        2,
        3,
        7,
        9,
        pytest.param(42, marks=pytest.mark.heavy_2a),
        pytest.param(97, marks=pytest.mark.heavy_2a),
        pytest.param(100, marks=pytest.mark.heavy_2a),
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        pytest.param(1, marks=pytest.mark.heavy_3b),
        2,
        5,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        pytest.param(
            it.chain(
                range(10**4),
                np.logspace(10, 32, num=50, base=2, dtype="int"),
            ),
            marks=pytest.mark.heavy_3c,
        ),
        (i for i in range(10) for __ in range(2)),
        (10 - i for i in range(10) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**2,
            size=10,
        ),
        pytest.param(
            np.random.default_rng(1).integers(
                low=0,
                high=2**32,
                size=10,
            ),
            marks=pytest.mark.heavy_2b,
        ),
    ],
)
def test_ranks_sorted_and_unique(degree, interspersal, time_sequence):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec)
    for num_strata_deposited in time_sequence:
        for which in (
            instance,
            geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec),
        ):
            assert all(
                i < j
                for i, j in pairwise(
                    which(
                        policy,
                        num_strata_deposited,
                    )
                )
            )


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        pytest.param(1, marks=pytest.mark.heavy_3a),
        2,
        3,
        7,
        9,
        pytest.param(42, marks=pytest.mark.heavy_2a),
        pytest.param(97, marks=pytest.mark.heavy_2a),
        pytest.param(100, marks=pytest.mark.heavy_2a),
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        pytest.param(1, marks=pytest.mark.heavy_3b),
        2,
        5,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        pytest.param(
            it.chain(
                range(10**4),
                np.logspace(10, 32, num=50, base=2, dtype="int"),
            ),
            marks=pytest.mark.heavy_3c,
        ),
        (i for i in range(10) for __ in range(2)),
        (10 - i for i in range(10) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**2,
            size=10,
        ),
        pytest.param(
            np.random.default_rng(1).integers(
                low=0,
                high=2**32,
                size=10,
            ),
            marks=pytest.mark.heavy_2b,
        ),
    ],
)
def test_zero_and_last_ranks_retained(degree, interspersal, time_sequence):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec)
    for num_strata_deposited in time_sequence:
        for which in (
            instance,
            geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec),
        ):
            res = which(
                policy,
                num_strata_deposited,
            )
            if num_strata_deposited > 1:
                first, *middle, last = res
                assert first == 0
                assert last == num_strata_deposited - 1
            elif num_strata_deposited == 1:
                assert ip.popsingleton(res) == 0
            else:
                assert next(res, None) is None


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        pytest.param(1, marks=pytest.mark.heavy_3a),
        2,
        3,
        7,
        9,
        pytest.param(42, marks=pytest.mark.heavy_2a),
        pytest.param(97, marks=pytest.mark.heavy_2a),
        pytest.param(100, marks=pytest.mark.heavy_2a),
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        pytest.param(1, marks=pytest.mark.heavy_3b),
        2,
        5,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        pytest.param(
            it.chain(
                range(10**4),
                np.logspace(10, 32, num=50, base=2, dtype="int"),
            ),
            marks=pytest.mark.heavy_3c,
        ),
        (i for i in range(10) for __ in range(2)),
        (10 - i for i in range(10) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=10**2,
            size=10,
        ),
        pytest.param(
            np.random.default_rng(1).integers(
                low=0,
                high=2**32,
                size=10,
            ),
            marks=pytest.mark.heavy_2b,
        ),
    ],
)
def test_ranks_valid(degree, interspersal, time_sequence):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec)
    for num_strata_deposited in time_sequence:
        for which in (
            instance,
            geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec),
        ):
            assert all(
                isinstance(r, numbers.Integral)
                and 0 <= r < num_strata_deposited
                for r in which(policy, num_strata_deposited)
            )


@pytest.mark.filterwarnings(
    "ignore:Interspersal set to 1, no bound on MRCA rank estimate uncertainty can be guaranteed."
)
@pytest.mark.parametrize(
    "degree",
    [
        1,
        2,
        3,
        7,
        9,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "interspersal",
    [
        1,
        2,
        5,
    ],
)
def test_eq(degree, interspersal):
    policy = geom_seq_nth_root_tapered_algo.Policy(degree, interspersal)
    spec = policy.GetSpec()
    instance = geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec)

    assert instance == instance
    assert instance == geom_seq_nth_root_tapered_algo.IterRetainedRanks(spec)
    assert instance is not None



import itertools as it
import unittest

from hstrat._auxiliary_lib import pairwise


class TestAllSame(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        for container in list, dict, set, iter:
            assert [*pairwise(container(()))] == []

    def test_singleton(self):
        for container in list, iter:
            for content in None, True, False, "a", 1, 0, "asdf", [1, 2, 3]:
                assert [*pairwise(container((content,)))] == []

    def test_different_types(self):
        for container in list, iter:
            content = [None, True, False, "a", [1, 2, 3]]
            assert [*pairwise(container(content))] == [
                (None, True),
                (True, False),
                (False, "a"),
                ("a", [1, 2, 3]),
            ]

    def test_doublet_same(self):
        for container in list, iter:
            for content in None, True, False, "a", 1, 0, "asdf", [1, 2, 3]:
                assert [*pairwise(container((content, content)))] == [
                    (content, content),
                ]

    def test_doublet_notsame(self):
        content = None, True, False, "a", 1, 0, "asdf", [1, 2, 3]
        for container in list, iter:
            for a, b in it.permutations(content, 2):
                assert [*pairwise(container((a, b)))] == [
                    (a, b),
                ]

    def test_triplet_same(self):
        for container in list, iter:
            for content in None, True, False, "a", 1, 0, "asdf", [1, 2, 3]:
                assert [*pairwise(container((content, content, content)))] == [
                    (content, content),
                    (content, content),
                ]

    def test_same_type(self):
        for container in list, iter:
            content = range(4)
            assert [*pairwise(container(content))] == [
                (0, 1),
                (1, 2),
                (2, 3),
            ]


if __name__ == "__main__":
    unittest.main()



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import alifestd_has_contiguous_ids

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")[-1:0],
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        )[0:1],
    ],
)
def test_alifestd_has_contiguous_ids_true(phylogeny_df):
    phylogeny_df_ = phylogeny_df.copy()
    assert alifestd_has_contiguous_ids(phylogeny_df)
    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_has_contiguous_ids_false(phylogeny_df):
    phylogeny_df_ = phylogeny_df.copy()
    assert not alifestd_has_contiguous_ids(phylogeny_df)
    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_ranks_since_first_retained_disparity_with import (
    calc_ranks_since_first_retained_disparity_with,
)


def calc_definitive_min_ranks_since_first_retained_disparity_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
) -> typing.Optional[int]:
    """Determine a hard, exclusive lower bound on generations since MRCA.

    At least, how many depositions have elapsed along the focal column's line
    of descent since the first mismatching strata at the same rank between
    focal and other?

    Returns None if no disparity found (i.e., both columns have same number
    of strata deposited and the most recent stratum is common between focal
    and other).
    """
    confidence_level = 0.49
    assert (
        focal.GetStratumDifferentiaBitWidth()
        == other.GetStratumDifferentiaBitWidth()
    )
    assert (
        calc_min_implausible_spurious_consecutive_differentia_collisions_between(
            focal,
            other,
            significance_level=1.0 - confidence_level,
        )
        == 1
    )
    return calc_ranks_since_first_retained_disparity_with(
        focal,
        other,
        confidence_level=confidence_level,
    )



import pandas as pd

from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_make_ancestor_id_col import alifestd_make_ancestor_id_col
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._all_unique import all_unique
from ._is_subset import is_subset


def _validate_ancestors_asexual(
    phylogeny_df: pd.DataFrame, mutate: bool
) -> bool:
    if "ancestor_id" not in phylogeny_df:
        if not mutate:
            phylogeny_df = phylogeny_df.copy()
        phylogeny_df["ancestor_id"] = alifestd_make_ancestor_id_col(
            phylogeny_df["id"], phylogeny_df["ancestor_list"]
        )
    elif not (
        phylogeny_df["ancestor_list"]
        .astype("str")
        .str.lower()
        .replace("[]", "[none]")
        == alifestd_make_ancestor_list_col(
            phylogeny_df["id"], phylogeny_df["ancestor_id"]
        )
    ).all():
        return False

    return is_subset(
        phylogeny_df["ancestor_id"].to_numpy(),
        phylogeny_df["id"].astype("int").to_numpy(),
    )


def _validate_ancestors_sexual(phylogeny_df: pd.DataFrame) -> bool:
    ids = set(phylogeny_df["id"])
    return all(
        ancestor_id in ids
        for ancestor_list_str in phylogeny_df["ancestor_list"]
        for ancestor_id in alifestd_parse_ancestor_ids(ancestor_list_str)
        if ancestor_id is not None
    )


def _alifestd_validate(
    phylogeny_df: pd.DataFrame,
    mutate: bool,
) -> bool:
    has_mandatory_columns = (
        "id" in phylogeny_df and "ancestor_list" in phylogeny_df
    )
    if not has_mandatory_columns:
        return False

    ids_valid = (
        len(phylogeny_df) == 0
        or pd.api.types.is_integer_dtype(phylogeny_df["id"])
        and (phylogeny_df["id"] >= 0).all()
        and all_unique(phylogeny_df["id"].to_numpy())
    )
    if not ids_valid:
        return False

    ancestor_lists_syntax_valid = (
        len(phylogeny_df) == 0
        or pd.api.types.is_string_dtype(phylogeny_df["ancestor_list"])
        and phylogeny_df["ancestor_list"].str.startswith("[").all()
        and phylogeny_df["ancestor_list"].str.endswith("]").all()
    )
    if not ancestor_lists_syntax_valid:
        return False

    return (
        _validate_ancestors_asexual(phylogeny_df, mutate)
        if alifestd_is_asexual(phylogeny_df)
        else _validate_ancestors_sexual(phylogeny_df)
    )


def alifestd_validate(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> bool:
    """Is the phylogeny compliant to alife data standards?

    Input dataframe is not mutated by this operation unless `mutate` set True.
    """

    try:
        return _alifestd_validate(phylogeny_df, mutate)
    except:
        return False



import typing_extensions

from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._impl import DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH
from ._pack_differentiae_bytes import pack_differentiae_bytes


def col_to_packet(
    column: HereditaryStratigraphicColumn,
    num_strata_deposited_byte_width: int = (
        DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH
    ),
) -> typing_extensions.Buffer:
    """Serialize a `HereditaryStratigraphicColumn` to a dict composed of
    builtin types."""

    if (
        not column._CanOmitStratumDepositionRank()
        or column.HasAnyAnnotations()
    ):
        raise NotImplementedError()

    total_differentia_bits = (
        column.GetNumStrataRetained() * column.GetStratumDifferentiaBitWidth()
    )
    total_differentia_bytes = (total_differentia_bits + 7) // 8  # +7 rounds up
    needs_num_padding_bits_header: bool = (
        bool(column.GetStratumDifferentiaBitWidth() % 8)
        and column._stratum_retention_policy.CalcNumStrataRetainedExact is None
    )
    if needs_num_padding_bits_header:
        # extra byte to encode num padding bits
        total_differentia_bytes += 1

    total_packet_bytes = (
        num_strata_deposited_byte_width + total_differentia_bytes
    )

    packet_buffer = bytearray(total_packet_bytes)
    assert len(packet_buffer) == total_packet_bytes
    packet_buffer[
        :num_strata_deposited_byte_width
    ] = column.GetNumStrataDeposited().to_bytes(
        num_strata_deposited_byte_width,
        byteorder="big",
        signed=False,
    )
    assert len(packet_buffer) == total_packet_bytes
    differentia_bytes = pack_differentiae_bytes(
        column.IterRetainedStrata(),
        column.GetStratumDifferentiaBitWidth(),
        always_omit_num_padding_bits_header=not needs_num_padding_bits_header,
    )
    assert len(differentia_bytes) == total_differentia_bytes, len(
        differentia_bytes
    )
    assert len(packet_buffer[num_strata_deposited_byte_width:]) == len(
        differentia_bytes
    )
    packet_buffer[num_strata_deposited_byte_width:] = differentia_bytes
    assert len(packet_buffer) == total_packet_bytes
    return packet_buffer



import typing


def alifestd_parse_ancestor_ids(ancestor_list_str: str) -> typing.List[int]:
    """Parse ancestor ids from an `ancestor_list` field."""
    if ancestor_list_str.lower() == "[none]":
        return []
    else:
        return eval(ancestor_list_str)



import mmh3


def decide_if_discard(
    hash_salt: int,
    stratum_rank: int,
    num_stratum_depositions_completed: int,
) -> bool:
    return (
        mmh3.hash(
            str(
                (
                    stratum_rank,
                    num_stratum_depositions_completed,
                )
            ),
            hash_salt,
        )
        % 2
    )



import copy
import itertools as it
import random

import anytree
import contexttimer as ctt
import pytest

from hstrat._auxiliary_lib import anytree_iterative_deepcopy


@pytest.fixture
def singleton_tree():
    root = anytree.Node("A")
    return root


@pytest.fixture
def linked_list_tree():
    root = anytree.Node("A")
    b = anytree.Node("B", parent=root)
    c = anytree.Node("C", parent=b)
    anytree.Node("D", parent=c)
    return root


@pytest.fixture
def multifurcating_tree():
    root = anytree.Node("A")
    b = anytree.Node("B", parent=root)
    c = anytree.Node("C", parent=root)
    d = anytree.Node("D", parent=root)
    anytree.Node("E", parent=b)
    anytree.Node("F", parent=c)
    anytree.Node("G", parent=c)
    anytree.Node("H", parent=d)
    anytree.Node("I", parent=d)
    anytree.Node("J", parent=d)
    return root


@pytest.fixture
def large_tree():
    max_depth = 50

    def create_tree(root, depth):
        if depth >= max_depth:
            return
        for i in range(random.choice((1, 1, 1, 1, 1, 1, 1, 2))):
            child = anytree.Node(f"{i}", parent=root)
            create_tree(child, depth + 1)

    # create the root node
    root = anytree.Node("root")
    # create tree from root
    create_tree(root, 0)
    return root


@pytest.mark.parametrize(
    "tree_fixture",
    [
        "singleton_tree",
        "linked_list_tree",
        "multifurcating_tree",
        "large_tree",
    ],
)
def test_iterative_deepcopy(tree_fixture, request):
    tree = request.getfixturevalue(tree_fixture)
    for subtree in it.islice(anytree.PreOrderIter(tree), 10):
        subtree_snapshot = str(anytree.RenderTree(subtree).by_attr("name"))

        subtree_copy = anytree_iterative_deepcopy(subtree)
        subtree_copy_snapshot = str(
            anytree.RenderTree(subtree_copy).by_attr("name")
        )
        assert subtree_copy_snapshot == subtree_snapshot

        subtree_after_snapshot = str(
            anytree.RenderTree(subtree).by_attr("name")
        )
        assert subtree_after_snapshot == subtree_snapshot


def test_benchmark(large_tree):
    with ctt.Timer(factor=1000) as t_recursive:
        for __ in range(1):
            copy.deepcopy(large_tree)

    with ctt.Timer(factor=1000) as t_iterative:
        for __ in range(1):
            anytree_iterative_deepcopy(large_tree)

    print(f"t_recursive={t_recursive}")
    print(f"t_iterative={t_iterative} ")



from copy import copy
import math
import operator
import sys
import typing
import warnings

from deprecated.sphinx import deprecated
from interval_search import binary_search
import opytional as opyt

from ..stratum_retention_strategy.stratum_retention_algorithms import (
    perfect_resolution_algo,
)
from ._HereditaryStratum import HereditaryStratum
from .stratum_ordered_stores import HereditaryStratumOrderedStoreList
from .stratum_ordered_stores._detail import HereditaryStratumOrderedStoreBase

# define type alias for ordered stores
OrderedStore = typing.Union[
    typing.Callable[..., HereditaryStratumOrderedStoreBase],
    typing.Tuple[HereditaryStratumOrderedStoreBase, int],
    None,
]


class HereditaryStratigraphicColumn:
    """Genetic annotation to enable phylogenetic inference.

    Primary end-user facing interface for hstrat library. Should be bundled with
    digital genomes and propagated via the CloneDescendant method when passing
    those genomes from parent to offspring. Provides basis for phylogenetic
    analysis of distributed digital evolution populations.

    Naming conventions are derived by analogy to Geological "Stratigraphy"
    (i.e., <https://en.wikipedia.org/wiki/Stratigraphy>). The "hereditary
    stratigraphy" system provided by this software works by associating an
    identifier, referred to as a "stratum," with each elapsed generation along
    a line of descent. This allows two "columns" to be aligned to detect the
    generation of their most recent common ancestor: strata before the MRCA
    will be identical and strata after will differ.

    Stratum retention policy and stratum differentia bit width can be configured
    to tune the time and space complexity of the column, trading-off with
    uncertainty induced on estimates of phylogenetic distance back to the most
    common recent ancestor of two columns.

    Arbitrary user-defined data can be associated with strata by optional
    argument to the CloneDescendant method. (Note that a first stratum is
    deposited during column initialization, so an optional annotation argument
    may also be provided then.)
    """

    __slots__ = (
        "_always_store_rank_in_stratum",
        "_stratum_differentia_bit_width",
        "_num_strata_deposited",
        "_stratum_ordered_store",
        "_stratum_retention_policy",
    )

    # if True, strata will be constructed with deposition rank stored even if
    # the stratum retention condemner does not require it
    _always_store_rank_in_stratum: bool
    # how many bits wide of differentia should the deposited strata be
    # constructed with?
    _stratum_differentia_bit_width: int
    # counter tracking the number of strata deposited
    # incremented *after* a deposition and its coinciding purge are complete
    _num_strata_deposited: int
    # data structure storing retained strata
    _stratum_ordered_store: typing.Any
    # functor specifying stratum retention policy
    _stratum_retention_policy: typing.Any

    def __init__(
        self: "HereditaryStratigraphicColumn",
        stratum_retention_policy: typing.Any = perfect_resolution_algo.Policy(),
        *,
        always_store_rank_in_stratum: bool = True,
        stratum_differentia_bit_width: int = 64,
        initial_stratum_annotation: typing.Optional[typing.Any] = None,
        stratum_ordered_store: OrderedStore = None,
        stratum_ordered_store_factory: OrderedStore = None  # deprecated
    ):
        """Initialize column to track a new line of descent.

        Deposits a first stratum, so GetNumStrataDeposited() will return 1 after
        initialization even though the user has not yet called DepositStratum().

        Parameters
        ----------
        stratum_retention_policy : any
            Policy struct that implements stratum retention policy by specifying
            the set of strata ranks that should be pruned from a hereditary
            stratigraphic column when the nth stratum is deposited.
        always_store_rank_in_stratum : bool, optional
            Should the deposition rank be stored as a data member of generated
            strata, even if not strictly necessary?
        stratum_differentia_bit_width : int, optional
            The bit width of the generated differentia. Default 64, allowing
            for 2^64 distinct values.
        initial_stratum_annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with the first stratum deposition in the
            line of descent.
        stratum_ordered_store: callable or tuple of store and count, optional
            One of:
            * callable to generate a container that implements the necessary
            interface to store strata within the column; can be configured for
            performance reasons, but has no semantic effect.
            * instance of one aforementioned container along with a deposition count
            * None, in which case a default-initialized container will be used
        stratum_ordered_store_factory: deprecated, alias of stratum_ordered_store.

        Notes
        -----
        If no condemner or predicate functor specifying a stratum retention
        policy is provided, the perfect resolution policy where all strata are
        retained is used.
        """
        self._always_store_rank_in_stratum = always_store_rank_in_stratum
        self._stratum_differentia_bit_width = stratum_differentia_bit_width
        self._stratum_retention_policy = stratum_retention_policy

        if stratum_ordered_store_factory is not None:
            warnings.warn(
                """stratum_ordered_store_factory kwarg is deprecated.
                Please use stratum_ordered_store kwarg instead.""",
                DeprecationWarning,
            )
            # disallow mixed use of deprecated and replacement
            assert stratum_ordered_store is None
            stratum_ordered_store = stratum_ordered_store_factory

        if stratum_ordered_store is None:
            # if no hstrat ordered store is specified, we use a list
            stratum_ordered_store = HereditaryStratumOrderedStoreList
        if callable(stratum_ordered_store):
            # ordered store is actually an ordered store factory
            self._stratum_ordered_store = stratum_ordered_store()
            self._num_strata_deposited = 0
            self.DepositStratum(annotation=initial_stratum_annotation)
        elif isinstance(
            stratum_ordered_store[0], HereditaryStratumOrderedStoreBase
        ):
            # ordered store is already an instance of an ordered store
            (
                self._stratum_ordered_store,
                self._num_strata_deposited,
            ) = stratum_ordered_store
        else:
            raise ValueError(
                """stratum_ordered_store is of invalid type; \
            should be callable or tuple(callable instance, deposition count)"""
            )

    def __eq__(
        self: "HereditaryStratigraphicColumn",
        other: "HereditaryStratigraphicColumn",
    ) -> bool:
        """Compare for value-wise equality."""
        # adapted from https://stackoverflow.com/a/4522896
        return (
            isinstance(
                other,
                self.__class__,
            )
            and self.__slots__ == other.__slots__
            and all(
                getter(self) == getter(other)
                for getter in [
                    operator.attrgetter(attr) for attr in self.__slots__
                ]
            )
        )

    def _CanOmitStratumDepositionRank(
        self: "HereditaryStratigraphicColumn",
    ) -> bool:
        """Does the column's associated stratum retention policy provide
        calculation of the rank of a stratum as a function of its position?"""
        return self._stratum_retention_policy.CalcRankAtColumnIndex is not None

    def _ShouldOmitStratumDepositionRank(
        self: "HereditaryStratigraphicColumn",
    ) -> bool:
        """Decide if deposition rank should be stored in stratum.

        Implementation detail to inspect configured stratum retention policy
        and manual override to decide whether deposition rank should be stored
        as a data member of generated strata.

        Note that strata are not required to be stored if the stratum retention
        policy allows for the rank of a stratum to be calcualted as a function
        of its position within the column and the number of strata deposited
        onto the column. However, it may be beneficial to store the stratum
        anyways for performance reasons if this calculation is expenxive.
        """
        return (
            self._CanOmitStratumDepositionRank()
            and not self._always_store_rank_in_stratum
        )

    def _CreateStratum(
        self: "HereditaryStratigraphicColumn",
        deposition_rank: int,
        annotation: typing.Optional[typing.Any] = None,
        differentia: typing.Optional[int] = None,
    ) -> HereditaryStratum:
        """Create a hereditary stratum with stored configuration attributes."""
        return HereditaryStratum(
            annotation=annotation,
            deposition_rank=(
                # don't store deposition rank if we know how to
                # calcualte it from stratum's position in column
                None
                if self._ShouldOmitStratumDepositionRank()
                else deposition_rank
            ),
            differentia_bit_width=self._stratum_differentia_bit_width,
            differentia=differentia,
        )

    def DepositStratum(
        self: "HereditaryStratigraphicColumn",
        annotation: typing.Optional[typing.Any] = None,
    ) -> None:
        """Elapse a generation.

        Parameters
        ----------
        annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum deposition in the
            line of descent.
        """
        new_stratum = self._CreateStratum(
            deposition_rank=self._num_strata_deposited,
            annotation=annotation,
        )
        self._stratum_ordered_store.DepositStratum(
            rank=self._num_strata_deposited,
            stratum=new_stratum,
        )
        self._PurgeColumn()
        self._num_strata_deposited += 1

    def DepositStrata(
        self: "HereditaryStratigraphicColumn",
        num_stratum_depositions: int,
    ) -> None:
        """Elapse n generations.

        Parameters
        ----------
        num_stratum_depositions: int
            How many generations to elapse?
        """
        # fallback to naive approach if IterRetainedRanks not available
        policy = self._stratum_retention_policy
        if policy.IterRetainedRanks is None or num_stratum_depositions <= 1:
            for __ in range(num_stratum_depositions):
                self.DepositStratum()
            return

        prior_deposit_count = self.GetNumStrataDeposited()
        target_deposit_count = prior_deposit_count + num_stratum_depositions

        # for python 3.7+, dictionaries are guaranteed insertion ordered
        assert sys.version_info >= (3, 7)
        target_retained_ranks = {
            rank: None
            for rank in policy.IterRetainedRanks(target_deposit_count)
        }

        # delete no-longer-needed ranks
        self._stratum_ordered_store.DelRanks(
            ranks=(
                rank
                for rank in self.IterRetainedRanks()
                if rank not in target_retained_ranks
            ),
            get_column_index_of_rank=self.GetColumnIndexOfRank,
        )

        # add new retained ranks
        for rank in target_retained_ranks:
            if rank >= prior_deposit_count:
                self._stratum_ordered_store.DepositStratum(
                    rank=rank,
                    stratum=self._CreateStratum(rank),
                )

        # update generation counter
        self._num_strata_deposited += num_stratum_depositions

    def _PurgeColumn(self: "HereditaryStratigraphicColumn") -> None:
        """Discard stored strata according to the configured retention policy.

        Implementation detail. Called after a new stratum has been appended to
        the column's store but before it is considered fully deposited (i.e.,
        it is reflected in the column's internal deposition counter).
        """
        condemned_ranks = self._stratum_retention_policy.GenDropRanks(
            num_stratum_depositions_completed=self.GetNumStrataDeposited(),
            retained_ranks=self.IterRetainedRanks(),
        )
        self._stratum_ordered_store.DelRanks(
            ranks=condemned_ranks,
            get_column_index_of_rank=self.GetColumnIndexOfRank,
        )

    def IterRetainedRanks(
        self: "HereditaryStratigraphicColumn",
    ) -> typing.Iterator[int]:
        """Iterate over deposition ranks of strata stored in the column.

        The store may be altered during iteration without iterator
        invalidation, although subsequent updates will not be reflected in the
        iterator.
        """
        if self._ShouldOmitStratumDepositionRank():
            if hasattr(self._stratum_retention_policy, "IterRetainedRanks"):
                yield from self._stratum_retention_policy.IterRetainedRanks(
                    self.GetNumStrataDeposited()
                )
            else:
                for idx in range(self.GetNumStrataRetained()):
                    yield self.GetRankAtColumnIndex(idx)
        else:
            yield from self._stratum_ordered_store.IterRetainedRanks()

    def IterRetainedStrata(
        self: "HereditaryStratigraphicColumn",
    ) -> typing.Iterator[HereditaryStratum]:
        """Iterate over strata stored in the column.
        Strata yielded from most ancient to most recent.
        """
        yield from self._stratum_ordered_store.IterRetainedStrata()

    def IterRetainedDifferentia(
        self: "HereditaryStratigraphicColumn",
    ) -> typing.Iterator[int]:
        """Iterate over differentia of strata stored in the column.

        Differentia yielded from most ancient to most recent.
        """
        return map(lambda x: x.GetDifferentia(), self.IterRetainedStrata())

    def IterRankDifferentiaZip(
        self: "HereditaryStratigraphicColumn",
        copyable: bool = False,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over ranks of retained strata and their differentia.

        If `copyable`, return an iterator that can be copied to produce a new
        fully-independent iterator at the same position.

        Equivalent to `zip(col.IterRetainedRanks(),
        col.IterRetainedDifferentia())`, but may be more efficient.
        """
        res = self._stratum_ordered_store.IterRankDifferentiaZip()
        if copyable:
            res = iter([*res])
        return res

    def HasAnyAnnotations(
        self: "HereditaryStratigraphicColumn",
    ) -> bool:
        """Do any retained strata have annotations?"""
        return any(
            stratum.GetAnnotation() is not None
            for stratum in self._stratum_ordered_store.IterRetainedStrata()
        )

    def GetNumStrataRetained(self: "HereditaryStratigraphicColumn") -> int:
        """How many strata are currently stored within the column?

        May be fewer than the number of strata deposited if strata have been
        discarded as part of the configured stratum retention policy.
        """
        return self._stratum_ordered_store.GetNumStrataRetained()

    def GetNumStrataDeposited(self: "HereditaryStratigraphicColumn") -> int:
        """How many strata have been depostited on the column?

        Note that a first stratum is deposited on the column during
        initialization.
        """
        return self._num_strata_deposited

    def GetStratumAtColumnIndex(
        self: "HereditaryStratigraphicColumn",
        index: int,
    ) -> HereditaryStratum:
        """Get the stratum positioned at index i among retained strata.

        Index order is from most ancient (index 0) to most recent.
        """
        return self._stratum_ordered_store.GetStratumAtColumnIndex(
            index,
            get_rank_at_column_index=(
                self.GetRankAtColumnIndex
                if self._ShouldOmitStratumDepositionRank()
                else None
            ),
        )

    def GetStratumAtRank(
        self: "HereditaryStratigraphicColumn",
        rank: int,
    ) -> typing.Optional[HereditaryStratum]:
        """Get the stratum deposited at generation g.

        Returns None if stratum with rank g is not retained.
        """
        return opyt.apply_if(
            self.GetColumnIndexOfRank(rank),
            self.GetStratumAtColumnIndex,
        )

    def GetRankAtColumnIndex(
        self: "HereditaryStratigraphicColumn",
        index: int,
    ) -> int:
        """Map column position to generation of deposition.

        What is the deposition rank of the stratum positioned at index i
        among retained strata? Index order is from most ancient (index 0) to
        most recent.
        """
        if self._ShouldOmitStratumDepositionRank():
            return self._stratum_retention_policy.CalcRankAtColumnIndex(
                index=index,
                num_strata_deposited=self.GetNumStrataDeposited(),
            )
        else:
            # fall back to store lookup
            return self._stratum_ordered_store.GetRankAtColumnIndex(index)

    def GetColumnIndexOfRank(
        self: "HereditaryStratigraphicColumn",
        rank: int,
    ) -> typing.Optional[int]:
        """Map generation of deposition to column position.

        What is the index position within retained strata of the stratum
        deposited at rank r? Returns None if no stratum with rank r is present
        within the store.
        """
        if self._ShouldOmitStratumDepositionRank():
            assert self.GetNumStrataRetained()
            res_idx = binary_search(
                lambda idx: self.GetRankAtColumnIndex(idx) >= rank,
                0,
                self.GetNumStrataRetained() - 1,
            )
            if res_idx is None:
                return None
            elif self.GetRankAtColumnIndex(res_idx) == rank:
                return res_idx
            else:
                return None
        else:
            # fall back to store lookup
            return self._stratum_ordered_store.GetColumnIndexOfRank(rank=rank)

    def GetNumDiscardedStrata(
        self: "HereditaryStratigraphicColumn",
    ) -> int:
        """How many deposited strata have been discarded?

        Determined by number of generations elapsed and the configured column
        retention policy.
        """
        return self.GetNumStrataDeposited() - self.GetNumStrataRetained()

    def GetStratumDifferentiaBitWidth(
        self: "HereditaryStratigraphicColumn",
    ) -> int:
        """How many bits wide are the differentia of strata?"""
        return self._stratum_differentia_bit_width

    def HasDiscardedStrata(
        self: "HereditaryStratigraphicColumn",
    ) -> bool:
        """Have any deposited strata been discarded?"""
        return self.GetNumDiscardedStrata() > 0

    @deprecated(
        version="1.6.0",
        reason="Use calc_probability_differentia_collision_between",
    )
    def CalcProbabilityDifferentiaCollision(
        self: "HereditaryStratigraphicColumn",
    ) -> float:
        """How likely are differentia collisions?

        Calculates the probability of two randomly-differentiated differentia
        being identical by coincidence.
        """
        return 1.0 / 2**self._stratum_differentia_bit_width

    @deprecated(
        version="1.6.0",
        reason="Use "
        "calc_min_implausible_spurious_differentia_collisions_between",
    )
    def CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
        self: "HereditaryStratigraphicColumn",
        significance_level: float,
    ) -> int:
        """Determine amount of evidence required to indicate shared ancestry.

        Calculates how many differentia collisions are required to reject the
        null hypothesis that columns do not share common ancestry at those
        ranks at significance level significance_level.
        """
        assert 0.0 <= significance_level <= 1.0

        log_base = self.CalcProbabilityDifferentiaCollision()
        return int(math.ceil(math.log(significance_level, log_base)))

    def Clone(
        self: "HereditaryStratigraphicColumn",
    ) -> "HereditaryStratigraphicColumn":
        """Create an independent copy of the column.

        Contains identical data but may be freely altered without affecting
        data within this column.
        """
        # shallow copy
        result = copy(self)
        # do semi-shallow duplication on select elements
        result._stratum_ordered_store = self._stratum_ordered_store.Clone()
        return result

    def CloneDescendant(
        self: "HereditaryStratigraphicColumn",
        stratum_annotation: typing.Optional[typing.Any] = None,
    ) -> "HereditaryStratigraphicColumn":
        """Return a cloned column that has had an additional stratum deposited.

        Does not alter self.

        Parameters
        ----------
        stratum_annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum deposition in the
            line of descent.
        """
        res = self.Clone()
        res.DepositStratum(annotation=stratum_annotation)
        return res

    def CloneNthDescendant(
        self: "HereditaryStratigraphicColumn",
        num_stratum_depositions: int,
    ) -> "HereditaryStratigraphicColumn":
        """Return a cloned column that has had n additional strata deposited.

        Does not alter self.

        Parameters
        ----------
        num_stratum_depositions: int
            How many generations should clone column be descended?
        """
        res = self.Clone()
        res.DepositStrata(num_stratum_depositions)
        return res



import itertools as it
import typing


# adapted from https://stackoverflow.com/a/19564384
def unpairwise(
    iterable: typing.Iterable[typing.Tuple[typing.Any, typing.Any]]
) -> typing.Iterable:
    """Return an iterator that yields the first element of the first pair in
    iterable and then the second element of all pairs in iterable, effectively
    reversing a `pairwise` operation.

    Parameters
    ----------
    iterable : iterable
        A sequence of overlapping pairs to be "flattened".

    Returns
    -------
    iterator
        An iterator that interleaves the second element of each pair with the
        remaining elements in the input sequence.

    Notes
    -----
    This function is the precise inverse of pairwise, except in the singleton
    case when pairwise yields empty, effectively discarding the lone element.

    Examples
    --------
    >>> a = [(1, 2), (2, 3), (3, 4)]
    >>> list(unpairwise(a))
    [1, 2, 3, 4]
    """

    p = iter(iterable)
    return it.chain(next(p, []), (x[1] for x in p))



import typing

import pandas as pd

from hstrat._auxiliary_lib import (
    alifestd_assign_contiguous_ids,
    alifestd_find_chronological_inconsistency,
    alifestd_make_empty,
    alifestd_parse_ancestor_ids,
    alifestd_try_add_ancestor_id_col,
)


def _is_chronological_inconsistency(
    df: pd.DataFrame, id_: typing.Optional[int]
) -> bool:
    if id_ is None:
        return False
    ancestor_list_str = df.loc[df["id"] == id_, "ancestor_list"].iloc[0]
    ancestor_ids = alifestd_parse_ancestor_ids(ancestor_list_str)

    if not ancestor_ids:
        return False

    for parent_id in ancestor_ids:
        parent_origin_time = df.loc[df["id"] == parent_id, "origin_time"].iloc[
            0
        ]
        if parent_origin_time > df.loc[df["id"] == id_, "origin_time"].iloc[0]:
            return True

    return False


def test_find_chronological_inconsistency_empty():
    df = alifestd_make_empty()
    df["origin_time"] = []
    assert alifestd_find_chronological_inconsistency(df) is None


def test_is_chronologically_ordered_empty():
    df = alifestd_make_empty()
    df["origin_time"] = []
    assert alifestd_find_chronological_inconsistency(df) is None


def test_is_chronologically_ordered_sexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 0],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_find_chronological_inconsistency(df) is None
            assert (
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                )
                is None
            )
            assert df.equals(df_)


def test_is_chronologically_ordered_sexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 2],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_find_chronological_inconsistency(df) is None
            assert (
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                )
                is None
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_sexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 6],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(df),
            )
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                ),
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_sexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 1, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(df),
            )
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                ),
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_sexual3():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 3, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(df),
            )
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                ),
            )
            assert df.equals(df_)


def test_is_chronologically_ordered_asexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[5]",
                "[333]",
                "[none]",
                "[1]",
                "[333]",
            ],
            "origin_time": [6, 5, 2, 1, 6, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_find_chronological_inconsistency(df) is None
            assert (
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                )
                is None
            )
            assert df.equals(df_)


def test_is_chronologically_ordered_asexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[5]",
                "[333]",
                "[none]",
                "[1]",
                "[none]",
            ],
            "origin_time": [6, 5, 2, 1, 6, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert alifestd_find_chronological_inconsistency(df) is None
            assert (
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                )
                is None
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_asexual1():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333]",
                "[333]",
                "[333]",
                "[none]",
                "[333]",
                "[0]",
            ],
            "origin_time": [5, 5, 2, 1, 2, 4],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(df),
            )
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                ),
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_asexual2():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[0]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 1, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(df),
            )
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                ),
            )
            assert df.equals(df_)


def test_is_chronologically_unordered_asexual3():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333]",
                "[5]",
                "[44]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [1, 5, 2, 3, 2, 2],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            df_ = df.copy()
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(df),
            )
            assert _is_chronological_inconsistency(
                df_,
                alifestd_find_chronological_inconsistency(
                    df.astype({"origin_time": float})
                ),
            )
            assert df.equals(df_)



import random
import typing

import pytest

from hstrat._auxiliary_lib import RngStateContext


@pytest.fixture
def seed_values() -> typing.Dict[int, typing.List[float]]:
    """Returns a dictionary of seed values to their first five elements.

    Returns:
    --------
    seed_values: Dict[int, List[float]]
        A dictionary mapping seed values to a list of their first five elements.
    """
    seed_list = [10, 20, 30, 42, 50, 420, 12345]
    num_elements = 5
    seed_values = {}
    for seed in seed_list:
        random.seed(seed)
        seed_values[seed] = [random.random() for _ in range(num_elements)]
    return seed_values


def test_reseeded_rng_temporarily_reseed_rng(seed_values):
    outer_seed = 42
    inner_seed = 420

    random.seed(outer_seed)
    assert random.random() == seed_values[outer_seed][0]
    with RngStateContext(inner_seed):
        for i, __ in enumerate(seed_values[inner_seed]):
            assert random.random() == seed_values[inner_seed][i]
    assert random.random() == seed_values[outer_seed][1]


def test_reseeded_rng_temporarily_reseed_rng_multiple_times(seed_values):
    seed1 = 10
    expected_values1 = seed_values[seed1]
    with RngStateContext(seed1):
        for i, __ in enumerate(expected_values1):
            assert random.random() == expected_values1[i]

    seed2 = 20
    expected_values2 = seed_values[seed2]
    with RngStateContext(seed2):
        for i, __ in enumerate(expected_values2):
            assert random.random() == expected_values2[i]

    seed3 = 30
    expected_values3 = seed_values[seed3]
    with RngStateContext(seed3):
        for i, __ in enumerate(expected_values3):
            assert random.random() == expected_values3[i]

    assert random.random() != expected_values3[-1]
    assert random.random() != expected_values2[-1]
    assert random.random() != expected_values1[-1]


def test_reseeded_rng_nested_reseeding(seed_values):
    outer_seed = 42
    inner_seed = 420

    random.seed(outer_seed)
    assert random.random() == seed_values[outer_seed][0]

    with RngStateContext(inner_seed):
        assert random.random() == seed_values[inner_seed][0]

        inner_inner_seed = 12345
        with RngStateContext(inner_inner_seed):
            for j, __ in enumerate(seed_values[inner_inner_seed]):
                assert random.random() == seed_values[inner_inner_seed][j]

        assert random.random() == seed_values[inner_seed][1]

    assert random.random() == seed_values[outer_seed][1]


def test_reseeded_rng_exception(seed_values):
    seed = 50
    expected_value = seed_values[seed][0]
    with pytest.raises(ValueError):
        with RngStateContext(seed):
            assert random.random() == expected_value
            raise ValueError()
    assert random.random() != expected_value



import pytest

from hstrat import hstrat


@pytest.fixture(
    params=[
        policy_t(
            parameterizer=hstrat.PropertyExactlyParameterizer(
                target_value=target_value,
                policy_evaluator=hstrat.MrcaUncertaintyAbsExactEvaluator(
                    at_num_strata_deposited=256,
                    at_rank=0,
                ),
                param_lower_bound=lb,
                param_upper_bound=1024,
            )
        )
        for policy_t, lb in (
            (hstrat.fixed_resolution_algo.Policy, 1),
            (hstrat.depth_proportional_resolution_algo.Policy, 1),
            (hstrat.depth_proportional_resolution_tapered_algo.Policy, 1),
            (hstrat.recency_proportional_resolution_algo.Policy, 0),
        )
        for target_value in (31, 127)
    ]
    + [
        policy_t(
            parameterizer=hstrat.PropertyAtLeastParameterizer(
                target_value=31,
                policy_evaluator=hstrat.MrcaUncertaintyAbsExactEvaluator(
                    at_num_strata_deposited=256,
                    at_rank=0,
                ),
                param_lower_bound=1,
                param_upper_bound=1024,
            )
        )
        for policy_t in (
            hstrat.geom_seq_nth_root_algo.Policy,
            hstrat.geom_seq_nth_root_tapered_algo.Policy,
            hstrat.recency_proportional_resolution_curbed_algo.Policy,
        )
    ]
    + [
        hstrat.geom_seq_nth_root_algo.Policy(
            parameterizer=hstrat.PropertyExactlyParameterizer(
                target_value=127,
                policy_evaluator=hstrat.MrcaUncertaintyAbsExactEvaluator(
                    at_num_strata_deposited=256,
                    at_rank=0,
                ),
                param_lower_bound=1,
                param_upper_bound=1024,
            )
        )
    ]
    + [
        hstrat.recency_proportional_resolution_curbed_algo.Policy(
            parameterizer=hstrat.PropertyExactlyParameterizer(
                target_value=127,
                policy_evaluator=hstrat.MrcaUncertaintyAbsExactEvaluator(
                    at_num_strata_deposited=256,
                    at_rank=0,
                ),
                param_lower_bound=1,
                param_upper_bound=1024,
            )
        )
    ]
    + [
        hstrat.recency_proportional_resolution_curbed_algo.Policy(
            size_curb=8,
        )
    ]
    + [
        hstrat.geom_seq_nth_root_tapered_algo.Policy(
            parameterizer=hstrat.PropertyAtMostParameterizer(
                target_value=127,
                policy_evaluator=hstrat.MrcaUncertaintyAbsExactEvaluator(
                    at_num_strata_deposited=256,
                    at_rank=0,
                ),
                param_lower_bound=1,
                param_upper_bound=1024,
            )
        )
    ],
)
def docplot_policy(request):
    return request.param



import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_does_share_any_common_ancestor(retention_policy, ordered_store):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    first_copy = first.Clone()
    second = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    def assert_validity():
        wrap = hstrat.col_to_specimen
        assert hstrat.does_share_any_common_ancestor(
            [wrap(first), wrap(first), wrap(first)]
        )
        assert hstrat.does_share_any_common_ancestor([first, first, first])
        assert hstrat.does_share_any_common_ancestor(
            [first_copy, first_copy, first_copy]
        )
        assert hstrat.does_share_any_common_ancestor(
            [first, first_copy, first]
        )
        assert hstrat.does_share_any_common_ancestor(
            [first_copy, first, first]
        )
        assert hstrat.does_share_any_common_ancestor(
            [first_copy, first, first] * 3
        )

        assert not hstrat.does_share_any_common_ancestor(
            [second, first, first]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [first, second, first]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [second, first_copy, first]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [first_copy, second, first]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [second, first, second]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [first, second, second]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [second, first_copy, second]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [first_copy, second, second] * 3
        )
        assert not hstrat.does_share_any_common_ancestor(
            [wrap(second), wrap(first_copy), wrap(second)]
        )
        assert not hstrat.does_share_any_common_ancestor(
            [wrap(first_copy), wrap(second), wrap(second)] * 3
        )

        assert hstrat.does_share_any_common_ancestor([first, first])
        assert hstrat.does_share_any_common_ancestor(
            [wrap(first), wrap(first)]
        )
        assert hstrat.does_share_any_common_ancestor([first_copy, first_copy])
        assert hstrat.does_share_any_common_ancestor([first, first_copy])
        assert hstrat.does_share_any_common_ancestor([first_copy, first])

        assert not hstrat.does_share_any_common_ancestor([second, first])
        assert not hstrat.does_share_any_common_ancestor([first, second])
        assert not hstrat.does_share_any_common_ancestor([second, first_copy])
        assert not hstrat.does_share_any_common_ancestor([first_copy, second])
        assert not hstrat.does_share_any_common_ancestor(
            [wrap(first_copy), wrap(second)]
        )

    assert_validity()
    first.DepositStratum()
    assert_validity()
    second.DepositStratum()
    assert_validity()
    first.DepositStratum()
    assert_validity()

    for __ in range(100):
        first_copy.DepositStratum()
        assert_validity()


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_does_share_any_common_ancestor_narrow(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    assert hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)]) is None
    assert hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)]) is None
    assert hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)]) is None
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2), wrap(c1)])
        is None
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1), wrap(c1)])
        is None
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)] * 3) is None
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)] * 3) is None
    )

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    assert hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)]) is False
    assert hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)]) is False
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2), wrap(c1)])
        is False
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1), wrap(c1)])
        is False
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)] * 3)
        is False
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)] * 3)
        is False
    )

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    c2 = c1.Clone()
    assert hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)]) is None
    assert hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)]) is None
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2), wrap(c1)])
        is None
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1), wrap(c1)])
        is None
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)] * 3) is None
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)] * 3) is None
    )

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    c2 = c1.Clone()
    assert hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)]) is True
    assert hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)]) is True
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2), wrap(c1)])
        is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1), wrap(c1)])
        is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)] * 3) is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)] * 3) is True
    )

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(100):
        c1.DepositStratum()
    c2 = c1.CloneDescendant()
    assert hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)]) is True
    assert hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)]) is True
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2), wrap(c1)])
        is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1), wrap(c1)])
        is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)] * 3) is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)] * 3) is True
    )

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(100):
        c1.DepositStratum()
    c2 = c1.CloneDescendant()
    assert hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)]) is True
    assert hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)]) is True
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2), wrap(c1)])
        is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1), wrap(c1)])
        is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c1), wrap(c2)] * 3) is True
    )
    assert (
        hstrat.does_share_any_common_ancestor([wrap(c2), wrap(c1)] * 3) is True
    )


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_does_share_any_common_ancestor_singleton(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(10):
        assert hstrat.does_share_any_common_ancestor([wrap(c1)]) is None
        c1.DepositStratum()

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(10):
        assert hstrat.does_share_any_common_ancestor([wrap(c1)]) is None
        c2.DepositStratum()


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
def test_does_share_any_common_ancestor_empty():
    assert hstrat.does_share_any_common_ancestor([]) is None


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_does_share_any_common_ancestor_generator(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(10):
        assert hstrat.does_share_any_common_ancestor(
            [wrap(c1) for __ in range(10)]
        ) == hstrat.does_share_any_common_ancestor(
            (wrap(c1) for __ in range(10))
        )
        c1.DepositStratum()

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(10):
        assert hstrat.does_share_any_common_ancestor(
            [wrap(c2) for __ in range(10)]
        ) == hstrat.does_share_any_common_ancestor(
            (wrap(c2) for __ in range(10))
        )
        c2.DepositStratum()



import itertools as it
import numbers

from iterpop import iterpop as ip
import numpy as np
import pytest

from hstrat._auxiliary_lib import all_same, pairwise
from hstrat.hstrat import recency_proportional_resolution_algo


@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_impl_consistency(recency_proportional_resolution, time_sequence):
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    spec = policy.GetSpec()
    impls = [
        *recency_proportional_resolution_algo._scry._IterRetainedRanks_.impls
    ]
    instances = [impl(spec) for impl in impls] + [
        lambda __, num_strata_deposited: policy.IterRetainedRanks(
            num_strata_deposited
        )
    ]
    for num_strata_deposited in time_sequence:
        assert all_same(
            it.chain(
                (
                    list(
                        impl(spec)(
                            policy,
                            num_strata_deposited,
                        )
                    )
                    for impl in impls
                ),
                (
                    list(
                        instance(
                            policy,
                            num_strata_deposited,
                        )
                    )
                    for instance in instances
                ),
            )
        )


@pytest.mark.parametrize(
    "impl",
    recency_proportional_resolution_algo._scry._IterRetainedRanks_.impls,
)
@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_only_dwindling_over_time(
    impl, recency_proportional_resolution, time_sequence
):
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = impl(spec)
    for num_strata_deposited in time_sequence:
        for which in (instance, impl(spec)):
            cur_set = {
                *which(
                    policy,
                    num_strata_deposited,
                )
            }
            next_set = {
                *which(
                    policy,
                    num_strata_deposited + 1,
                )
            }
            assert cur_set.issuperset(next_set - {num_strata_deposited})


@pytest.mark.parametrize(
    "impl",
    recency_proportional_resolution_algo._scry._IterRetainedRanks_.impls,
)
@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_ranks_sorted_and_unique(
    impl, recency_proportional_resolution, time_sequence
):
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = impl(spec)
    for num_strata_deposited in time_sequence:
        for which in (instance, impl(spec)):
            assert all(
                i < j
                for i, j in pairwise(
                    which(
                        policy,
                        num_strata_deposited,
                    )
                )
            )


@pytest.mark.parametrize(
    "impl",
    recency_proportional_resolution_algo._scry._IterRetainedRanks_.impls,
)
@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_zero_and_last_ranks_retained(
    impl, recency_proportional_resolution, time_sequence
):
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = impl(spec)
    for num_strata_deposited in time_sequence:
        for which in instance, impl(spec):
            res = which(
                policy,
                num_strata_deposited,
            )
            if num_strata_deposited > 1:
                first, *middle, last = res
                assert first == 0
                assert last == num_strata_deposited - 1
            elif num_strata_deposited == 1:
                assert ip.popsingleton(res) == 0
            else:
                assert next(res, None) is None


@pytest.mark.parametrize(
    "impl",
    recency_proportional_resolution_algo._scry._IterRetainedRanks_.impls,
)
@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_ranks_valid(impl, recency_proportional_resolution, time_sequence):
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = impl(spec)
    for num_strata_deposited in time_sequence:
        for which in (instance, impl(spec)):
            assert all(
                isinstance(r, numbers.Integral)
                and 0 <= r < num_strata_deposited
                for r in which(policy, num_strata_deposited)
            )


@pytest.mark.parametrize(
    "impl",
    recency_proportional_resolution_algo._scry._IterRetainedRanks_.impls,
)
@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(impl, recency_proportional_resolution):
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    spec = policy.GetSpec()
    instance = impl(spec)

    assert instance == instance
    assert instance == impl(spec)
    assert instance is not None



import itertools as it
import typing

from hstrat import _auxiliary_lib as auxlib
from hstrat.stratum_retention_strategy.stratum_retention_algorithms.recency_proportional_resolution_curbed_algo._impl import (
    calc_geom_seq_nth_root_transition_rank,
    calc_provided_resolution,
)


def iter_backing_policy_transition_ranks(
    size_curb: int,
) -> typing.Iterator[int]:

    gsnra_rank = calc_geom_seq_nth_root_transition_rank(size_curb)
    for i in it.count():
        cur_rank = 2**i
        assert auxlib.popcount(cur_rank) == 1

        if cur_rank >= gsnra_rank:
            if gsnra_rank:
                yield gsnra_rank
            return

        if cur_rank and calc_provided_resolution(
            size_curb, cur_rank - 1
        ) != calc_provided_resolution(size_curb, cur_rank):
            yield cur_rank
        else:
            continue

    assert False



import typing

import pandas as pd
import sortedcontainers as sc

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_mark_num_descendants_asexual import (
    alifestd_mark_num_descendants_asexual,
)
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_find_mrca_id_asexual(
    phylogeny_df: pd.DataFrame,
    leaf_ids: typing.Iterable[int],
    mutate: bool = False,
) -> int:
    """Find most recent common ancestor of `leaf_ids`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if "num_descendants" not in phylogeny_df:
        phylogeny_df = alifestd_mark_num_descendants_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    lineages = sc.SortedSet({*leaf_ids}, key=phylogeny_df.index.get_loc)
    if not len(lineages):
        raise ValueError()

    while len(lineages) > 1:
        oldest = lineages.pop(-1)
        replacement = phylogeny_df.loc[oldest, "ancestor_id"]
        assert replacement != oldest
        lineages.add(replacement)

    (mrca_id,) = lineages
    return mrca_id



import typing


def is_in(value: object, iterator: typing.Iterator) -> bool:
    """Check if a value is in an iterator using object identity comparison.

    Parameters
    ----------
    value : object
        The value to check for in the iterator.
    iterator : typing.Iterator
        An iterator of objects to check against the value.

    Returns
    -------
    bool
        True if the value is found in the iterator, False otherwise.

    Notes
    -----
    This function performs object identity comparison (using the `is` operator),
    not equality comparison (using the `==` operator).
    """
    for candidate in iterator:
        if value is candidate:
            return True
    return False



import typing

import numpy as np


def _apply_mutation(
    pop_arr: np.array,
    mut_distn: typing.Callable,
) -> None:
    """Apply mutation to all organisms' genetic trait, altering `pop_arr`
    inplace.
    """
    pop_arr += mut_distn(size=len(pop_arr))



import pytest

from hstrat._auxiliary_lib import raises


def test_raises_success():
    def my_func():
        raise ValueError("test")

    assert raises(my_func, ValueError) == True


def test_raises_failure():
    def my_func():
        raise ValueError("test")

    with pytest.raises(ValueError):
        assert raises(my_func, TypeError) == False


def test_raises_no_exception():
    def my_func():
        return True

    assert raises(my_func, ValueError) == False



import itertools as it
import typing

from bitstring import BitArray
import numpy as np
import typing_extensions

from .._auxiliary_lib import iter_chunks


def _numpy_unpack_differentiae(
    buffer: bytes, differentia_bit_width: int
) -> typing.Iterator[int]:
    # numpy dtypes can be composed of various attributes
    # in this case, we want our type to be big-endian (>),
    # and to be composed of `num_bytes` unsigned ints (u).
    # for more info, see https://numpy.org/doc/stable/reference/generated/numpy.dtype.html
    num_bytes = differentia_bit_width // 8
    dt = np.dtype(f">u{num_bytes}")
    yield from np.frombuffer(buffer, dtype=dt)


def _bitarray_unpack_differentiae(
    buffer: bytes,
    differentia_bit_width: int,
    num_packed_differentia: typing.Optional[int],
) -> typing.Iterator[int]:
    bits = BitArray(bytes=buffer)
    num_header_bits = (
        8
        if num_packed_differentia is None and bool(differentia_bit_width % 8)
        else 0
    )

    if num_packed_differentia is None:
        if len(bits) and num_header_bits:
            num_padding_bits = bits[:num_header_bits].uint
        else:
            num_padding_bits = 0
        num_valid_bits = len(bits) - num_header_bits - num_padding_bits
        assert num_valid_bits % differentia_bit_width == 0
        num_packed_differentia = num_valid_bits // differentia_bit_width

    valid_bits = bits[num_header_bits:]
    assert (
        len([*iter_chunks(valid_bits, differentia_bit_width)])
        >= num_packed_differentia
    )
    for chunk in it.islice(
        iter_chunks(valid_bits, differentia_bit_width), num_packed_differentia
    ):
        yield chunk.uint


def unpack_differentiae_bytes(
    packed_differentiae: typing_extensions.Buffer,
    differentia_bit_width: int,
    num_packed_differentia: typing.Optional[int] = None,
) -> typing.Iterator[int]:
    """Unpack a compact, concatenated byte buffer representation into
    a sequence with each element represented as a distinct integer."""
    if differentia_bit_width in [8, 16, 32, 64]:
        # bit width is a multiple of 8 -- no padding required
        assert (
            num_packed_differentia is None
            or num_packed_differentia * differentia_bit_width
            == len(packed_differentiae) * 8
        )
        yield from _numpy_unpack_differentiae(
            packed_differentiae, differentia_bit_width
        )
    else:
        # padding bits are possible, first byte tells how many are required
        yield from _bitarray_unpack_differentiae(
            packed_differentiae, differentia_bit_width, num_packed_differentia
        )



import numpy as np

from hstrat._auxiliary_lib import cmp


def test_cmp_greater_than():
    assert cmp(2, 1) == 1


def test_cmp_less_than():
    assert cmp(1, 2) == -1


def test_cmp_equal():
    assert cmp(2, 2) == 0


def test_cmp_strings():
    assert cmp("a", "b") == -1


def test_cmp_lists():
    assert cmp([1, 2, 3], [1, 2, 4]) == -1


def test_cmp_greater_than_numpy():
    assert cmp(np.double(2), np.double(1)) == 1


def test_cmp_less_than_numpy():
    assert cmp(np.double(1), np.double(2)) == -1


def test_cmp_equal_numpy():
    assert cmp(np.double(2), np.double(2)) == 0



import numpy as np

from ._jit import jit


@jit(nopython=True)
def indices_of_unique(array: np.array) -> np.array:
    """Return the indices where the first occurance of each unique value in
    `array` occurs.

    Equivalent to `return_index` functionality of `numpy.unique`, but faster.
    See <https://numpy.org/doc/1.24/reference/generated/numpy.unique.html>.
    """
    unique_indices = []
    seen_values = set()
    last_value = 0  # if seen_values empty, last_value not compared against
    for index, value in enumerate(array):
        is_same_as_last_value = len(unique_indices) and last_value == value
        if not is_same_as_last_value and value not in seen_values:
            unique_indices.append(index)
            seen_values.add(value)
            last_value = value
    return np.array(unique_indices)



import typing

import numpy as np

from ..._auxiliary_lib import iter_monotonic_equivalencies, jit
from ...frozen_instrumentation import HereditaryStratigraphicSpecimen


@jit(nopython=True)
def _compare_differentia_at_common_ranks(
    first_ranks: np.array,
    first_differentiae: np.array,
    second_ranks: np.array,
    second_differentiae: np.array,
) -> typing.Iterator[typing.Tuple[typing.Tuple[int, int], bool]]:
    for pos1, pos2 in iter_monotonic_equivalencies(first_ranks, second_ranks):
        yield (
            (pos1, pos2),
            first_differentiae[pos1] == second_differentiae[pos2],
        )


def iter_mutual_rank_indices(
    first: HereditaryStratigraphicSpecimen,
    second: HereditaryStratigraphicSpecimen,
    compare: bool = False,
) -> typing.Union[
    typing.Iterator[typing.Tuple[int, int]],
    typing.Iterator[typing.Tuple[typing.Tuple[int, int], bool]],
]:
    """Iterate over ranks with matching strata between columns in ascending
    order."""
    if compare:
        return _compare_differentia_at_common_ranks(
            first.GetRankIndex(),
            first.GetDifferentiaVals(),
            second.GetRankIndex(),
            second.GetDifferentiaVals(),
        )

    else:
        return iter_monotonic_equivalencies(
            first.GetRankIndex(), second.GetRankIndex()
        )



from copy import copy
import itertools as it
import operator
import sys
import typing

from .._HereditaryStratum import HereditaryStratum
from ._detail import HereditaryStratumOrderedStoreBase


class HereditaryStratumOrderedStoreDict(HereditaryStratumOrderedStoreBase):
    """Interchangeable backing container for HereditaryStratigraphicColumn.

    Stores deposited strata using a dict implementation. Retained strata are
    stored as dict values with their associated deposition ranks as keys.

    Potentially useful in scenarios where large strata counts are retained or
    deleted strata tend to be more ancient.
    """

    __slots__ = ("_data",)

    # maps rank to stratum
    _data: typing.Dict[int, HereditaryStratum]

    def __init__(self: "HereditaryStratumOrderedStoreDict"):
        """Initialize instance variables."""
        self._data = {}

    def __eq__(
        self: "HereditaryStratumOrderedStoreDict",
        other: "HereditaryStratumOrderedStoreDict",
    ) -> bool:
        """Compare for value-wise equality."""
        # adapted from https://stackoverflow.com/a/4522896
        return (
            isinstance(
                other,
                self.__class__,
            )
            and self.__slots__ == other.__slots__
            and all(
                getter(self) == getter(other)
                for getter in [
                    operator.attrgetter(attr) for attr in self.__slots__
                ]
            )
        )

    def DepositStratum(
        self: "HereditaryStratumOrderedStoreDict",
        rank: typing.Optional[int],
        stratum: "HereditaryStratum",
    ) -> None:
        """Insert a new stratum into the store.

        Parameters
        ----------
        rank : typing.Optional[int]
            The position of the stratum being deposited within the sequence of strata deposited into the column. Precisely, the number of strata that have been deposited before stratum.
        stratum : HereditaryStratum
            The stratum to deposit.
        """
        self._data[rank] = stratum

    def GetNumStrataRetained(self: "HereditaryStratumOrderedStoreDict") -> int:
        """How many strata are present in the store?

        May be fewer than the number of strata deposited if deletions have
        occured.
        """
        return len(self._data)

    def GetStratumAtColumnIndex(
        self: "HereditaryStratumOrderedStoreDict",
        index: int,
        get_rank_at_column_index: typing.Optional[typing.Callable] = None,
    ) -> HereditaryStratum:
        """Get the stratum positioned at index i among retained strata.

        Index order is from most ancient (index 0) to most recent.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the index position within retained strata of
            the stratum deposited at rank r.
        """
        if get_rank_at_column_index is not None:
            return self._data[get_rank_at_column_index(index)]
        else:
            # for python 3.7+, dictionaries are guaranteed insertion ordered
            assert sys.version_info >= (3, 7)
            return next(it.islice(self._data.values(), index, None))

    def GetRankAtColumnIndex(
        self: "HereditaryStratumOrderedStoreDict",
        index: int,
    ) -> int:
        """Map deposition generation to column position.

        What is the deposition rank of the stratum positioned at index i
        among retained strata? Index order is from most ancient (index 0) to
        most recent.
        """
        # for python 3.7+, dictionaries are guaranteed insertion ordered
        assert sys.version_info >= (3, 7)
        return next(it.islice(self._data.keys(), index, None))

    def GetColumnIndexOfRank(
        self: "HereditaryStratumOrderedStoreDict",
        rank: int,
    ) -> typing.Optional[int]:
        """Map column position ot deposition generation.

        What is the index position within retained strata of the stratum
        deposited at rank r? Returns None if no stratum with rank r is present
        within the store.
        """
        # for python 3.7+, dictionaries are guaranteed insertion ordered
        assert sys.version_info >= (3, 7)
        try:
            return next(
                idx
                for idx, rank_ in enumerate(self._data.keys())
                if rank_ == rank
            )
        except StopIteration:
            return None

    def DelRanks(
        self: "HereditaryStratumOrderedStoreDict",
        ranks: typing.Iterable[int],
        # needed for other implementations
        get_column_index_of_rank: typing.Optional[typing.Callable] = None,
    ) -> None:
        """Purge strata with specified deposition ranks from the store.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the deposition rank of the stratum positioned
            at index i among retained strata. Not used in this method.
        """
        for rank in ranks:
            del self._data[rank]

    def IterRetainedRanks(
        self: "HereditaryStratumOrderedStoreDict",
    ) -> typing.Iterator[int]:
        """Iterate over deposition ranks of strata present in the store from
        most ancient to most recent.

        The store may be altered during iteration without iterator
        invalidation, although subsequent updates will not be reflected in the
        iterator.
        """
        # must make copy to prevent
        # `RuntimeError: dictionary changed size during iteration`
        # note, however, that copy is made lazily
        # (only when first item requested)
        yield from list(self._data.keys())

    def IterRetainedStrata(
        self: "HereditaryStratumOrderedStoreDict",
    ) -> typing.Iterator[HereditaryStratum]:
        """Iterate over stored strata from most ancient to most recent."""
        # for python 3.7+, dictionaries are guaranteed insertion ordered
        assert sys.version_info >= (3, 7)
        yield from self._data.values()

    def IterRankDifferentiaZip(
        self: "HereditaryStratumOrderedStoreDict",
        # needed for other implementations
        get_rank_at_column_index: typing.Optional[typing.Callable] = None,
        start_column_index: int = 0,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over differentia and corresponding deposition ranks.

        Values yielded as tuples. Guaranteed ordered from most ancient to most
        recent.

        Parameters
        ----------
        get_rank_at_column_index : callable, optional
            Callable that returns the deposition rank of the stratum positioned
            at index i among retained strata.
        start_column_index : callable, optional
            Number of strata to skip over before yielding first result from the
            iterator. Default 0, meaning no strata are skipped over.
        """
        # optimization idea:
        # python dicts are ordered, so is there a way to begin iterating
        # from a specified item in the dict?
        # current method must iterate past skipped over items
        # adapted from https://stackoverflow.com/a/12911454
        iter_ = it.islice(self._data.items(), start_column_index, None)
        for rank, stratum in iter_:
            yield (rank, stratum.GetDifferentia())

    def Clone(
        self: "HereditaryStratumOrderedStoreDict",
    ) -> "HereditaryStratumOrderedStoreDict":
        """Create an independent copy of the store.

        Returned copy contains identical data but may be freely altered without
        affecting data within this store.
        """
        # shallow copy
        result = copy(self)
        # do semi-shallow clone on select elements
        # see https://stackoverflow.com/a/5861653 for performance consierations
        result._data = self._data.copy()
        return result



"""Tools to infer phylogenetic history from extant genome annotations."""

from . import estimators, pairwise, population, priors, tree
from .estimators import *  # noqa: F401
from .pairwise import *  # noqa: F401
from .population import *  # noqa: F401
from .priors import *  # noqa: F401
from .tree import *  # noqa: F401

__all__ = (
    estimators.__all__
    + pairwise.__all__
    + population.__all__
    + priors.__all__
    + tree.__all__
)



import typing

from ...._detail import PolicyCouplerBase
from ..._PolicySpec import PolicySpec


class GenDropRanksNaive:
    """Functor to implement the recency-proportional resolution stratum
    retention policy, for use with HereditaryStratigraphicColumn.

    This functor enacts the recency-proportional resolution policy by
    specifying the set of strata ranks that should be purged from a hereditary
    stratigraphic column when the nth stratum is deposited.

    Calculation uses a naive approach where set subtraction between retained
    ranks (i.e., `IterRetainedRanks()` at timepoint n+1 and timepoint n
    provides the set of ranks to drop. This approach is inefficient, but
    provided for internal validation and testing.
    """

    def __init__(
        self: "GenDropRanksNaive",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "GenDropRanksNaive", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    @staticmethod
    def _do_call(
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        retained_ranks: typing.Optional[typing.Iterable[int]],
    ) -> typing.Iterator[int]:
        """Implementation for __call__ to faciliate external (but within-
        library) calls."""
        if num_stratum_depositions_completed < 2:
            return

        yield from sorted(
            set(policy.IterRetainedRanks(num_stratum_depositions_completed))
            - set(
                policy.IterRetainedRanks(num_stratum_depositions_completed + 1)
            )
        )

    def __call__(
        self: "GenDropRanksNaive",
        policy: PolicyCouplerBase,
        num_stratum_depositions_completed: int,
        retained_ranks: typing.Optional[typing.Iterable[int]],
    ) -> typing.Iterator[int]:
        """Decide which strata within the stratagraphic column should be purged.

        Every time a new stratum is deposited, this method is called to
        determine which strata should be purged. All strata at ranks yielded
        from this functor are immediately purged from the column, meaning that
        for a stratum to persist it must not be yielded by this functor each
        and every time a new stratum is deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.
        retained_ranks : iterator over int
            An iterator over ranks of strata currently retained within the
            hereditary stratigraphic column. Not used in this functor.

        Returns
        -------
        iterator over int
            The ranks of strata that should be purged from the hereditary
            stratigraphic column at this deposition step.

        See Also
        --------
        recency_proportional_resolution_algo:
            For details on the rationale, implementation, and guarantees of the
            recency-proportional resolution stratum retention policy.
        """
        return GenDropRanksNaive._do_call(
            policy,
            num_stratum_depositions_completed,
            retained_ranks,
        )



import operator

import numpy as np
import pandas as pd

from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._alifestd_unfurl_lineage_asexual import alifestd_unfurl_lineage_asexual


def alifestd_prune_extinct_lineages_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Drop taxa without extant descendants.

    An "extant" column, if provided, is used to determine extant taxa.
    Otherwise, taxa with inf or nan "destruction_time" are considered extant.

    Fastest with records in working format. See `alifestd_to_working_format`.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.
    mutate : bool, default False
        Are side effects on the input argument `phylogeny_df` allowed?

    Raises
    ------
    ValueError
        If `phylogeny_df` has neither "extant" or "destruction_time" columns.

        Without at least one of these columns, which taxa are extant is
        ambiguous.

    Returns
    -------
    pandas.DataFrame
        The rerooted phylogeny in alife standard format.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    phylogeny_df.set_index("id", drop=False, inplace=True)

    extant_mask = None
    if "extant" in phylogeny_df:
        extant_mask = phylogeny_df["extant"]
    elif "destruction_time" in phylogeny_df:
        extant_mask = operator.or_(
            phylogeny_df["destruction_time"].isna(),
            np.isinf(phylogeny_df["destruction_time"]),
        )
    else:
        raise ValueError('Need "extant" or "destruction_time" column.')

    phylogeny_df["has_extant_descendant"] = False

    for extant_id in phylogeny_df.loc[extant_mask, "id"]:
        for lineage_id in alifestd_unfurl_lineage_asexual(
            phylogeny_df,
            extant_id,
            mutate=True,
        ):
            if phylogeny_df.loc[lineage_id, "has_extant_descendant"]:
                break

            phylogeny_df.loc[lineage_id, "has_extant_descendant"] = True

    drop_filter = ~phylogeny_df["has_extant_descendant"]
    phylogeny_df.drop(
        phylogeny_df.index[drop_filter], inplace=True, axis="rows"
    )
    phylogeny_df.drop("has_extant_descendant", inplace=True, axis="columns")
    return phylogeny_df.reset_index(drop=True)



"""Logic to control stratum retention within HereditaryStratigraphicColumn.

Stratum retention policies specify which strata ranks that should be retained ----- and which should be purged --- when the nth stratum is deposited.
Stratum retention algorithms can be parameterized to yield a policy.
"""

from . import (
    depth_proportional_resolution_algo,
    depth_proportional_resolution_tapered_algo,
    fixed_resolution_algo,
    geom_seq_nth_root_algo,
    geom_seq_nth_root_tapered_algo,
    nominal_resolution_algo,
    perfect_resolution_algo,
    pseudostochastic_algo,
    recency_proportional_resolution_algo,
    recency_proportional_resolution_curbed_algo,
    stochastic_algo,
)
from ._detail import UnsatisfiableParameterizationRequestError

provided_stratum_retention_algorithms = [
    depth_proportional_resolution_algo,
    depth_proportional_resolution_tapered_algo,
    fixed_resolution_algo,
    geom_seq_nth_root_algo,
    geom_seq_nth_root_tapered_algo,
    nominal_resolution_algo,
    perfect_resolution_algo,
    pseudostochastic_algo,
    recency_proportional_resolution_algo,
    stochastic_algo,
]

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "UnsatisfiableParameterizationRequestError",
    "depth_proportional_resolution_algo",
    "depth_proportional_resolution_tapered_algo",
    "fixed_resolution_algo",
    "geom_seq_nth_root_algo",
    "geom_seq_nth_root_tapered_algo",
    "nominal_resolution_algo",
    "perfect_resolution_algo",
    "provided_stratum_retention_algorithms",
    "pseudostochastic_algo",
    "recency_proportional_resolution_algo",
    "recency_proportional_resolution_curbed_algo",
    "stochastic_algo",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import typing

from .._detail._PolicyCouplerBase import PolicyCouplerBase


class CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce:
    def __init__(
        self: "CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce",
        policy_spec: typing.Optional[typing.Any] = None,
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce",
        policy: typing.Optional[PolicyCouplerBase],
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
    ) -> int:
        """Calculate pessimal rank for upper bound on absolute MRCA
        uncertainty by brute force."""
        least_num_strata_deposited = min(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )

        if least_num_strata_deposited == 0:
            return 0

        pessimal_rank = max(
            (
                (
                    r,
                    policy.CalcMrcaUncertaintyAbsUpperBound(
                        first_num_strata_deposited,
                        second_num_strata_deposited,
                        r,
                    ),
                )
                for r in range(least_num_strata_deposited)
            ),
            key=lambda tup: tup[1],
        )[0]

        return pessimal_rank



from collections import Counter
import os

import alifedata_phyloinformatics_convert as apc
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_find_leaf_ids,
    alifestd_has_contiguous_ids,
    alifestd_is_asexual,
    alifestd_is_sexual,
    alifestd_is_topologically_sorted,
    alifestd_make_ancestor_id_col,
    alifestd_to_working_format,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df1",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")[-1:0],
    ],
)
@pytest.mark.parametrize(
    "phylogeny_df2",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")[-1:0],
        None,
    ],
)
@pytest.mark.parametrize(
    "phylogeny_df3",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")[-1:0],
        None,
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_try_add_ancestor_id_col,
        alifestd_to_working_format,
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "apply_all",
    [
        True,
        False,
    ],
)
def test_alifestd_aggregate_phylogenies(
    phylogeny_df1,
    phylogeny_df2,
    phylogeny_df3,
    apply,
    apply_all,
):

    if phylogeny_df3 is None and phylogeny_df2 is not None:
        return

    phylogeny_df1 = apply(phylogeny_df1)
    phylogeny_df1_ = phylogeny_df1.copy()
    phylogenies = [phylogeny_df1]
    if phylogeny_df2 is not None:
        if apply_all:
            phylogeny_df2 = apply(phylogeny_df2)
        phylogeny_df2_ = phylogeny_df2.copy()
        phylogenies.append(phylogeny_df2)
    if phylogeny_df3 is not None:
        if apply_all:
            phylogeny_df3 = apply(phylogeny_df3)
        phylogeny_df3_ = phylogeny_df3.copy()
        phylogenies.append(phylogeny_df3)

    aggregate_df = alifestd_aggregate_phylogenies(phylogenies)
    assert alifestd_validate(aggregate_df)
    # check for side effects
    assert phylogeny_df1.equals(phylogeny_df1_)
    if phylogeny_df2 is not None:
        assert phylogeny_df2.equals(phylogeny_df2_)
    if phylogeny_df3 is not None:
        assert phylogeny_df3.equals(phylogeny_df3_)

    assert len(alifestd_find_leaf_ids(aggregate_df)) == sum(
        map(len, map(alifestd_find_leaf_ids, phylogenies)),
    )
    assert all(
        map(alifestd_has_contiguous_ids, phylogenies)
    ) == alifestd_has_contiguous_ids(aggregate_df)
    assert all(map(alifestd_is_asexual, phylogenies)) == alifestd_is_asexual(
        aggregate_df
    )
    assert all(
        map(alifestd_is_topologically_sorted, phylogenies)
    ) == alifestd_is_topologically_sorted(aggregate_df)
    assert any(map(alifestd_is_sexual, phylogenies)) == alifestd_is_sexual(
        aggregate_df
    )

    assert len(aggregate_df["id"].unique()) == len(aggregate_df["id"])
    if "ancestor_id" in aggregate_df:
        assert (
            aggregate_df["ancestor_id"]
            == alifestd_make_ancestor_id_col(
                aggregate_df["id"], aggregate_df["ancestor_list"]
            )
        ).all()

    if alifestd_is_asexual(aggregate_df):
        assert len(apc.alife_dataframe_to_dendropy_trees(aggregate_df)) == sum(
            bool(len(phylogeny)) for phylogeny in phylogenies
        )
        assert Counter(
            (node.level(), len(node.child_nodes()))
            for tree in apc.alife_dataframe_to_dendropy_trees(aggregate_df)
            for node in tree
        ) == Counter(
            (node.level(), len(node.child_nodes()))
            for tree in map(apc.alife_dataframe_to_dendropy_tree, phylogenies)
            if tree is not None
            for node in tree
        )



import operator
import random
import typing


class HereditaryStratum:
    """Packages stratigraph data associated with a particular generation.

    Includes differentia "fingerprint" associated with a particular
    generation of a line of descent and possibly additional metadata, including
    optional user-provided annotation.
    """

    __slots__ = (
        "_deposition_rank",
        "_differentia",
        "_annotation",
    )

    _deposition_rank: int
    # random "fingerprint" generated at initialization
    _differentia: int
    # optional arbitrary user-provided data
    _annotation: typing.Optional[typing.Any]

    def __init__(
        self: "HereditaryStratum",
        *,
        annotation: typing.Optional[typing.Any] = None,
        differentia_bit_width: int = 64,
        deposition_rank: typing.Optional[int] = None,
        differentia: typing.Optional[int] = None,
    ):
        """Construct the stratum.

        Randomly generates and stores a differentia "fingerprint" alongside
        other metadata, if provided.

        Parameters
        ----------
        annotation: any, optional
            Optional object to store as an annotation. Allows arbitrary user-
            provided to be associated with this stratum's generation in its
            line of descent.
        differentia_bit_width: int, optional
            The bit width of the generated differentia. Default 64, allowing
            for 2^64 distinct values.
        deposition_rank : int, optional
            The position of the stratum being deposited within the sequence of strata deposited into the column. Precisely, the number of strata that have been deposited before stratum.
        """
        if differentia is not None:
            if differentia_bit_width is not None:
                assert differentia < 2**differentia_bit_width
            self._differentia = differentia
        else:
            self._differentia = random.randrange(2**differentia_bit_width)

        self._annotation = annotation
        self._deposition_rank = deposition_rank

    def __eq__(self: "HereditaryStratum", other: "HereditaryStratum") -> bool:
        """Compare for value-wise equality."""
        # adapted from https://stackoverflow.com/a/4522896
        return (
            isinstance(
                other,
                self.__class__,
            )
            and self.__slots__ == other.__slots__
            and all(
                getter(self) == getter(other)
                for getter in [
                    operator.attrgetter(attr) for attr in self.__slots__
                ]
            )
        )

    def GetDepositionRank(self: "HereditaryStratum") -> typing.Optional[int]:
        """Get the deposition order rank associated with this stratum, if stored.

        Deposition rank is the number of strata deposited on a column before
        self. Deposition rank may not be stored if the stratum retention policy
        supports calculation of deposition rank from column index.
        """
        if hasattr(self, "_deposition_rank"):
            return self._deposition_rank
        else:
            return None

    def GetDifferentia(self: "HereditaryStratum") -> int:
        """Access differentia.

        Returns the randomly-generated value that distinguishes this stratum
        from others generated at the same rank in other hereditary columns.
        """
        return self._differentia

    def GetAnnotation(
        self: "HereditaryStratum",
    ) -> typing.Optional[typing.Any]:
        """Access arbitrary, user-specified annotation, if any."""
        return self._annotation



"""Animations for stratum retention policies."""

from ._policy_panel_animate import policy_panel_animate
from ._stratum_retention_animate import stratum_retention_animate

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "policy_panel_animate",
    "stratum_retention_animate",
]

from ..._auxiliary_lib import launder_impl_modules as _launder

_launder([eval(item) for item in __all__], __name__)
del _launder  # prevent name from leaking



import unittest

from hstrat._auxiliary_lib import consume


class TestConsume(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_iterable(self):
        try:
            consume(range(100))
        except TypeError:
            pass
        else:
            assert False

        try:
            consume([1, 2, 3, 4])
        except TypeError:
            pass
        else:
            assert False

        for n in 0, 1:
            try:
                consume([], n)
            except TypeError:
                pass
            else:
                assert False

    def test_empty(self):
        x = iter([])
        consume(x)
        for __ in x:
            assert False

        for n in 0, 1:
            x = iter(range(0))
            consume(x, n)
            for __ in x:
                assert False

    def test_singleton(self):
        x = iter("a")
        consume(x)
        assert [*x] == []

        x = iter("a")
        consume(x, 0)
        assert [*x] == ["a"]

        for n in 1, 2, 3:
            x = iter(range(1))
            consume(x, n)
            assert [*x] == []

    def test_pair(self):
        x = iter("ab")
        consume(x)
        assert [*x] == []

        x = iter("ab")
        consume(x, 0)
        assert [*x] == ["a", "b"]

        x = iter(range(2))
        consume(x, 1)
        assert [*x] == [1]

        x = iter(range(2))
        consume(x, 2)
        assert [*x] == []

        x = iter(range(2))
        consume(x, 3)
        assert [*x] == []


if __name__ == "__main__":
    unittest.main()
